// $Id: Arguments.cpp,v 1.15 2000/03/27 16:49:20 chuang Exp $
#include "TclObject.h"
#include "naCmd.h"
#include "Arguments.h"

Arguments::Arguments ():
    m_args(0),
    m_outValues(0)
{ }

Arguments::~Arguments ()
{
    delete[] m_args;
    delete[] m_outValues;
}

int
Arguments::initArgument (Tcl_Interp *interp,
                         Tcl_Obj *pObj,
                         int j,
                         const Parameter &parameter)
{
    TclObject argument(pObj);
    VARTYPE vt = parameter.type().vartype();

    if (pObj->typePtr == &naObjType) {
        // This variant indicates a missing optional argument.
        m_args[j] = vtMissing;

    } else if (parameter.flags() & PARAMFLAG_FOUT) {
        // For out parameters, set a pointer to where the out value
        // will be stored.

        if (vt == VT_USERDEFINED) {
            // Assume user defined types derive from IUnknown.
            vt = VT_UNKNOWN;
        }

        m_args[j].vt = vt | VT_BYREF;
        if (vt == VT_VARIANT) {
            // Set a pointer to out variant.
            m_args[j].byref = &m_outValues[j];
        } else {
            // Set a pointer to variant data value.
            m_args[j].byref = &m_outValues[j].byref;
        }

        if (parameter.flags() & PARAMFLAG_FIN) {
            // Set the value for an in/out parameter.
            Tcl_Obj *pValue = Tcl_ObjGetVar2(
                interp,
                pObj,
                NULL,
                TCL_PARSE_PART1 | TCL_LEAVE_ERR_MSG);
            if (pValue == 0) {
                return TCL_ERROR;
            }

            TclObject value(pValue);
            m_outValues[j] = value.toVariant(vt, interp);
        } else {
            if (vt == VT_UNKNOWN) {
                m_outValues[j].vt = vt;
                m_outValues[j].punkVal = NULL;
            } else if (vt != VT_VARIANT) {
                m_outValues[j].ChangeType(vt);
            }
        }

    } else {
        m_args[j] = argument.toVariant(vt, interp);
    }

    return TCL_OK;
}

void
Arguments::storeOutValues (Tcl_Interp *interp,
                           int objc,
                           Tcl_Obj *CONST objv[],
                           const Method::Parameters &parameters)
{
    if (objc > 0) {
        int j = objc - 1;
        Method::Parameters::const_iterator p = parameters.begin();
        for (int i = 0; i < objc; ++i, --j, ++p) {
            if (p->flags() & PARAMFLAG_FOUT) {
                TclObject value(m_outValues[j], p->type(), interp);
                Tcl_ObjSetVar2(
                    interp,
                    objv[i],
                    NULL,
                    value,
                    TCL_PARSE_PART1 | TCL_LEAVE_ERR_MSG);
            }
        }
    }
}


int
PositionalArguments::init (Tcl_Interp *interp,
                           int objc,
                           Tcl_Obj *CONST objv[],
                           const Method &method,
                           INVOKEKIND invokeKind)
{
    const Method::Parameters &parameters = method.parameters();

    int m = parameters.size();
    int n = objc;
    if (invokeKind == INVOKE_PROPERTYPUT
     || invokeKind == INVOKE_PROPERTYPUTREF) {
        m = objc;
        --n;
    }

    if (m > 0) {
        m_args = new _variant_t[m];
        m_outValues = new _variant_t[m];

        int j = m - 1;
        Method::Parameters::const_iterator p = parameters.begin();

        // Convert the arguments actually provided.
        for (int i = 0; i < n; ++i, --j, ++p) {
            int result = initArgument(interp, objv[i], j, *p);
            if (result != TCL_OK) {
                return result;
            }
        }

        // Fill in missing arguments.
        for (; p != parameters.end(); ++p, --j) {
            m_args[j] = vtMissing;
        }

        // Convert argument for property put operations.
        if (invokeKind == INVOKE_PROPERTYPUT) {
            TclObject value = objv[i];
            m_args[j] = value.toVariant(method.type().vartype(), interp);

        } else if (invokeKind == INVOKE_PROPERTYPUTREF) {
            TclObject value = objv[i];
            VARTYPE vt = method.type().vartype();
            m_outValues[j] = value.toVariant(vt, interp);
            m_args[j].vt = vt | VT_BYREF;
            m_args[j].byref = &m_outValues[j].byref;
        }
    }

    m_dispParams.rgvarg = m_args;
    m_dispParams.rgdispidNamedArgs = NULL;
    m_dispParams.cArgs = m;
    m_dispParams.cNamedArgs = 0;

    if (invokeKind == INVOKE_PROPERTYPUT
     || invokeKind == INVOKE_PROPERTYPUTREF) {
        static DISPID mydispid = DISPID_PROPERTYPUT;
        m_dispParams.rgdispidNamedArgs = &mydispid;
        m_dispParams.cNamedArgs = 1;
    }

    return TCL_OK;
}


NamedArguments::~NamedArguments ()
{
    delete[] m_namedDispids;
}

Method::Parameters::const_iterator
NamedArguments::findParameter (const Method::Parameters &parameters,
                               const std::string &name,
                               DISPID &dispid)
{
    int i = 0;
    Method::Parameters::const_iterator p = parameters.begin();
    for (; p != parameters.end(); ++p, ++i) {
        if (p->name() == name) {
            dispid = i;
            break;
        }
    }
    return p;
}

int
NamedArguments::init (Tcl_Interp *interp,
                      int objc,
                      Tcl_Obj *CONST objv[],
                      const Method &method,
                      INVOKEKIND)
{
    const Method::Parameters &parameters = method.parameters();

    if (objc % 2 != 0) {
        Tcl_AppendResult(interp, "name value pairs required", NULL);
        return TCL_ERROR;
    }

    int cArgs = objc / 2;
    if (cArgs > 0) {
        m_args = new _variant_t[cArgs];
        m_outValues = new _variant_t[cArgs];
        m_namedDispids = new DISPID[cArgs];

        int j = cArgs - 1;
        for (int i = 0; i < objc; i += 2, --j) {
            char *name = Tcl_GetStringFromObj(objv[i], 0);
            Method::Parameters::const_iterator p = findParameter(
                parameters,
                name,
                m_namedDispids[j]);
            if (p == parameters.end()) {
                Tcl_AppendResult(interp, "unknown parameter ", name, NULL);
                return TCL_ERROR;
            }

            int result = initArgument(interp, objv[i+1], j, *p);
            if (result != TCL_OK) {
                return result;
            }
        }
    }

    m_dispParams.rgvarg = m_args;
    m_dispParams.rgdispidNamedArgs = m_namedDispids;
    m_dispParams.cArgs = cArgs;
    m_dispParams.cNamedArgs = cArgs;

    return TCL_OK;
}
