// $Id: ComObject.cpp,v 1.3 2000/04/25 22:34:52 chuang Exp $
#pragma warning(disable: 4786)
#include <stdarg.h>
#include <stdexcept>
#include "CommandProcessor.h"
#include "interfacePointer.h"
#include "naCmd.h"
#include "refCmd.h"
#include "ComModule.h"
#include "ComObject.h"

ComObject::ComObject (const Interface &interfaceDesc,
                      Tcl_ThreadId threadId,
                      Tcl_Interp *interp,
                      TclObject servant,
                      TclObject destructor,
                      bool isSink):
    m_refCount(0),
    m_interface(interfaceDesc),
    m_threadId(threadId),
    m_interp(interp),
    m_servant(servant),
    m_destructor(destructor),
    m_isSink(isSink)
{
    ComModule::instance().lock();

//    Tcl_Preserve(reinterpret_cast<ClientData>(m_interp));

    // Initialize virtual function index to method description map.
    const Interface::Methods &methods = m_interface.methods();
    for (Interface::Methods::const_iterator p = methods.begin();
     p != methods.end(); ++p) {
        m_vtblIndexToMethodMap.insert(VtblIndexToMethodMap::value_type(
            p->vtblIndex(), &(*p)));
    }

    if (m_interface.dispatchable()) {
        m_pVtbl = dispatchVtbl;

        // Initialize dispatch member ID to method description map.
        const Interface::Methods &methods = m_interface.methods();
        for (Interface::Methods::const_iterator p = methods.begin();
         p != methods.end(); ++p) {
            m_dispIdToMethodMap.insert(DispIdToMethodMap::value_type(
                p->memberid(), &(*p)));
        }
    } else {
        m_pVtbl = unknownVtbl;
    }
}

ComObject::~ComObject ()
{
    // Execute destructor Tcl command if defined.
    int length;
    Tcl_GetStringFromObj(m_destructor, &length);
    if (length > 0) {
        TclObject script(m_destructor);
        script.lappend(m_servant);

        TclObject result;
        CommandProcessor::instance().eval(
            m_threadId, m_interp, script, &result);
    }

    ComModule::instance().unlock();
//    Tcl_Release(reinterpret_cast<ClientData>(m_interp));
}

IUnknown *
ComObject::newInstance (const Interface &interfaceDesc,
                        Tcl_ThreadId threadId,
                        Tcl_Interp *interp,
                        TclObject servant,
                        TclObject destructor,
                        bool isSink)
{
    return reinterpret_cast<IUnknown *>(new ComObject(
        interfaceDesc,
        threadId,
        interp,
        servant,
        destructor,
        isSink));
}

// Implement IUnknown methods

STDMETHODIMP
ComObject::QueryInterface (ComObject *pThis, REFIID iid, void **ppvObj)
{
    if (IsEqualIID(iid, pThis->m_interface.iid())
     || IsEqualIID(iid, IID_IUnknown)
     || (pThis->m_interface.dispatchable() && IsEqualIID(iid, IID_IDispatch))) {
	*ppvObj = pThis;
        ComObject::AddRef(pThis);
	return S_OK;
    }

    *ppvObj = 0;
    return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG)
ComObject::AddRef (ComObject *pThis)
{
    return InterlockedIncrement(&pThis->m_refCount);
}

STDMETHODIMP_(ULONG)
ComObject::Release (ComObject *pThis)
{
    ULONG refCount = InterlockedDecrement(&pThis->m_refCount);
    if (refCount == 0) {
	delete pThis;
    }
    return refCount;
}

// Implement IDispatch methods

STDMETHODIMP
ComObject::GetTypeInfoCount (ComObject * /*pThis*/, UINT *pCount)
{
    *pCount = 1;
    return S_OK;
}

STDMETHODIMP
ComObject::GetTypeInfo (ComObject *pThis,
                        UINT index,
                        LCID /*lcid*/,
                        ITypeInfo **ppTypeInfo)
{
    if (index != 0) {
        *ppTypeInfo = 0;
        return E_INVALIDARG;
    }

    ITypeInfo *pTypeInfo = pThis->m_interface.typeInfo();
    pTypeInfo->AddRef();
    *ppTypeInfo = pTypeInfo;
    return S_OK;
}

STDMETHODIMP
ComObject::GetIDsOfNames (ComObject *pThis,
                          REFIID /*riid*/,
                          OLECHAR **rgszNames,
                          UINT cNames,
                          LCID /*lcid*/,
                          DISPID *rgDispId)
{
    ITypeInfo *pTypeInfo = pThis->m_interface.typeInfo();
    return pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
}

// Generate a name of a Tcl variable used to hold an argument out value.

static TclObject
getOutVariableName (const Parameter &param)
{
    return TclObject("::tcomArg" + param.name());
}

// Convert IDispatch argument to Tcl value.

static TclObject
getArgument (Tcl_ThreadId threadId,
                 Tcl_Interp *interp,
                 _variant_t &argVar,
                 const Parameter &param)
{
    if (argVar == vtMissing) {
        return newNaObj();

    } else if (param.flags() & PARAMFLAG_FOUT) {
        TclObject varName = getOutVariableName(param);

        if (param.flags() & PARAMFLAG_FIN) {
            // For in/out parameters, put the argument value in a
            // Tcl variable.
            TclObject value(argVar, param.type(), interp);

            // Construct a Tcl command to set the variable.
            TclObject script(Tcl_NewListObj(0, 0));
            script.lappend(Tcl_NewStringObj("set", -1));
            script.lappend(varName);
            script.lappend(value);

            CommandProcessor::instance().eval(threadId, interp, script);
        }
        return varName;

    } else {
        return TclObject(argVar, param.type(), interp);
    }
}

// Fill exception information structure.

static void
fillExcepInfo (EXCEPINFO *pExcepInfo,
               HRESULT hresult,
               const char *source,
               const char *description)
{
    if (pExcepInfo != 0) {
        memset(pExcepInfo, 0, sizeof(EXCEPINFO));
        pExcepInfo->scode = hresult;

        _bstr_t bstrSource(source);
        pExcepInfo->bstrSource = SysAllocString(bstrSource);

        if (description != 0) {
            _bstr_t bstrDescription(description);
            pExcepInfo->bstrDescription = SysAllocString(bstrDescription);
        }
    }
}

STDMETHODIMP
ComObject::Invoke (ComObject *pThis,
                   DISPID dispid,
                   REFIID /*riid*/,
                   LCID /*lcid*/,
                   WORD wFlags,
                   DISPPARAMS *pDispParams,
                   VARIANT *pVarResult,
                   EXCEPINFO *pExcepInfo,
                   UINT *pArgErr)
{
    // Get the method description for method being invoked.
    DispIdToMethodMap::const_iterator p =
        pThis->m_dispIdToMethodMap.find(dispid);
    if (p == pThis->m_dispIdToMethodMap.end()) {
        return DISP_E_MEMBERNOTFOUND;
    }
    const Method *pMethod = p->second;

    HRESULT hresult;

    try {
        // Get the method or property to invoke on the servant.
        std::string operation;
        if (wFlags & DISPATCH_METHOD) {
            operation = pMethod->name();

        } else if (wFlags & (DISPATCH_PROPERTYPUT | DISPATCH_PROPERTYPUTREF)) {
            operation = "_set_" + pMethod->name();

        } else if (wFlags & DISPATCH_PROPERTYGET) {
            operation = "_get_" + pMethod->name();

        } else {
            return DISP_E_MEMBERNOTFOUND;
        }

        // Construct Tcl command call.
        TclObject script(pThis->m_servant);
        script.lappend(
            Tcl_NewStringObj(const_cast<char *>(operation.c_str()), -1));

        // Set the argument error pointer in case we need to use it.
        UINT argErr;
        if (pArgErr == 0) {
            pArgErr = &argErr;
        }

        // Convert arguments to Tcl values.
        int numPositionalArgs = pDispParams->cArgs - pDispParams->cNamedArgs;

        // Convert positional arguments to Tcl values.
        const Method::Parameters &parameters = pMethod->parameters();
        for (int i = 0; i < numPositionalArgs; ++i) {
            const Parameter &param = parameters[i];

            // Append argument value.
            _variant_t argVar(
                pDispParams->rgvarg[pDispParams->cArgs - 1 - i]);
            script.lappend(
                getArgument(pThis->m_threadId, pThis->m_interp, argVar, param));
        }

        // Convert named arguments to Tcl list of argument names and values.
        if (pDispParams->cNamedArgs > 0) {
            TclObject namedArguments(Tcl_NewListObj(0, 0));

            for (; i < pDispParams->cArgs; ++i) {
                int j = pDispParams->cNamedArgs - 1 - (i - numPositionalArgs);
                const Parameter &param =
                    parameters[pDispParams->rgdispidNamedArgs[j]];

                // Append argument name.
                TclObject argName(param.name());
                namedArguments.lappend(argName);

                // Append argument value.
                _variant_t argVar(
                    pDispParams->rgvarg[pDispParams->cArgs - 1 - i]);
                namedArguments.lappend(
                    getArgument(
                        pThis->m_threadId, pThis->m_interp, argVar, param));
            }

            script.lappend(namedArguments);
        }

        // Call the Tcl command.
        TclObject result;
        int completionCode = CommandProcessor::instance().eval(
            pThis->m_threadId,
            pThis->m_interp,
            script,
            pThis->m_isSink ? 0 : &result);
        if (completionCode != TCL_OK) {
            fillExcepInfo(
                pExcepInfo,
                E_UNEXPECTED,
                pThis->m_servant.c_str(),
                result.c_str());
            return DISP_E_EXCEPTION;
        }

        // Get values of out parameters.
        i = 0;
        Method::Parameters::const_iterator p = parameters.begin();
        for (; p != parameters.end(); ++p, ++i) {
            if (p->flags() & PARAMFLAG_FOUT) {
                TclObject varName = getOutVariableName(*p);

                // Construct a Tcl command to get the variable value.
                TclObject script(Tcl_NewListObj(0, 0));
                script.lappend(Tcl_NewStringObj("set", -1));
                script.lappend(varName);

                TclObject value;
                int completionCode = CommandProcessor::instance().eval(
                    pThis->m_threadId, pThis->m_interp, script, &value);

                if (completionCode == TCL_OK) {
                    pDispParams->rgvarg[pDispParams->cArgs - 1 - i] =
                        value.toVariant(p->type().vartype(), 0, true);
                }
            }
        }

        // Convert return value.
        VARTYPE vt = pMethod->type().vartype();
        if (vt != VT_VOID) {
            *pVarResult = result.toVariant(vt, 0, true);
        }

        hresult = S_OK;
    }
    catch (_com_error &e) {
        fillExcepInfo(pExcepInfo, e.Error(), pThis->m_servant.c_str(), 0);
        hresult = DISP_E_EXCEPTION;
    }
    return hresult;
}

static va_list
convertNativeToTclObject (va_list pArg,
                          Tcl_Interp *interp,
                          TclObject &tclObject,
                          const Type &type,
                          bool byRef=false)
{
    switch (type.vartype()) {
    case VT_BOOL:
        tclObject = Tcl_NewBooleanObj(
            byRef ? *va_arg(pArg, int *) : va_arg(pArg, int));
        break;

    case VT_R4:
    case VT_R8:
        tclObject = Tcl_NewDoubleObj(
            byRef ? *va_arg(pArg, double *) : va_arg(pArg, double));
        break;

    case VT_DATE:
        {
            double days =
                byRef ? *va_arg(pArg, double *) : va_arg(pArg, double);
            tclObject = Tcl_NewLongObj((days - 25570) * 86400);
        }
        break;

    case VT_DISPATCH:
    case VT_UNKNOWN:
    case VT_USERDEFINED:
        {
            IUnknown *pUnknown = va_arg(pArg, IUnknown *);
            const Interface *pInterface =
                InterfaceManager::instance().find(type.iid());
            tclObject = HandleSupport<Reference>::instance(interp).newObj(
                ReferenceFactory::instance().newReference(
                    pUnknown,
                    pInterface));
        }
        break;

    case VT_NULL:
        tclObject = Tcl_NewObj();
        break;

    case VT_BSTR:
        {
#if TCL_MINOR_VERSION >= 2
            // Uses Unicode function introduced in Tcl 8.2.
            tclObject = Tcl_NewUnicodeObj(va_arg(pArg, wchar_t *), -1);
#else
            tclObject = Tcl_NewStringObj(va_arg(pArg, char *), -1);
#endif
        }
        break;

    default:
        tclObject = Tcl_NewLongObj(
            byRef ? *va_arg(pArg, int *) : va_arg(pArg, int));
    }

    return pArg;
}

// Convert COM method argument to Tcl value.

static va_list
getArgument (Tcl_ThreadId threadId,
             Tcl_Interp *interp,
             va_list pArg,
             TclObject &tclObject,
             const Parameter &param)
{
    if (param.flags() & PARAMFLAG_FOUT) {
        TclObject varName = getOutVariableName(param);

        if (param.flags() & PARAMFLAG_FIN) {
            // For in/out parameters, put the argument value in a
            // Tcl variable.
            TclObject value;
            pArg = convertNativeToTclObject(
                pArg, interp, value, param.type(), true);

            // Construct a Tcl command to set the variable.
            TclObject script(Tcl_NewListObj(0, 0));
            script.lappend(Tcl_NewStringObj("set", -1));
            script.lappend(varName);
            script.lappend(value);

            CommandProcessor::instance().eval(threadId, interp, script);
        }
        tclObject = varName;
        return pArg;

    } else {
        TclObject value;
        return convertNativeToTclObject(
            pArg, interp, tclObject, param.type());
    }
}

// Convert Tcl value to out argument value.

static va_list
putArgument (va_list pArg,
             Tcl_Interp *interp,
             TclObject tclObject,
             const Type &type)
{
    switch (type.vartype()) {
    case VT_BOOL:
        *va_arg(pArg, int *) = tclObject.getBool() ? -1 : 0;
        break;

    case VT_R4:
        *va_arg(pArg, float *) = tclObject.getDouble();
        break;

    case VT_R8:
        *va_arg(pArg, double *) = tclObject.getDouble();
        break;

    case VT_DISPATCH:
    case VT_UNKNOWN:
    case VT_USERDEFINED:
        {
            IUnknown *pUnknown;

            Tcl_Obj *pObj = tclObject;
            if (pObj->typePtr == &interfacePointerObjType) {
                pUnknown = reinterpret_cast<IUnknown *>(
                    pObj->internalRep.otherValuePtr);
            } else {
                Reference *pRef =
                    HandleSupport<Reference>::instance(interp).find(tclObject);
                pUnknown = (pRef == 0) ? 0 : pRef->dispatch();
            }

            *va_arg(pArg, IUnknown **) = pUnknown;

            // The COM rules say we must increment the reference count of
            // interface pointers returned from methods.
            if (pUnknown != 0) {
                pUnknown->AddRef();
            }
        }
        break;

    case VT_BSTR:
        {
            BSTR *pBstr = va_arg(pArg, BSTR *);
#if TCL_MINOR_VERSION >= 2
            // Uses Unicode function introduced in Tcl 8.2.
            *pBstr = SysAllocString(tclObject.getUnicode());
#else
            *pBstr = SysAllocString(tclObject.c_str());
#endif
        }
        break;

    default:
        *va_arg(pArg, int *) = tclObject.getLong();
    }

    return pArg;
}

// Note that this function is called in an odd way to avoid copying the
// arguments onto the stack (for efficiency and simplicity in the calling
// code).  This is why the call is explicitly declared __cdecl.

void __cdecl
invokeComObjectFunction (volatile HRESULT hresult,
                         volatile DWORD pArgEnd,
	                 DWORD /*ebp*/,
	                 DWORD funcIndex,
	                 DWORD /*retAddr*/,
	                 ComObject *pThis,
                         ...)
{
    // Get the method description for method being invoked.
    ComObject::VtblIndexToMethodMap::const_iterator p =
        pThis->m_vtblIndexToMethodMap.find(funcIndex);
    if (p == pThis->m_vtblIndexToMethodMap.end()) {
        // If we don't have a method description, we don't know how many bytes
        // the arguments take on the stack.
        throw std::runtime_error("unknown virtual function index");
    }

    const Method *pMethod = p->second;

    // Construct Tcl command call.
    TclObject script(pThis->m_servant);

    std::string operation;
    switch (pMethod->invokeKind()) {
    case INVOKE_PROPERTYGET:
        operation = "_get_" + pMethod->name();
        break;
    case INVOKE_PROPERTYPUT:
    case INVOKE_PROPERTYPUTREF:
        operation = "_set_" + pMethod->name();
        break;
    default:
        operation = pMethod->name();
    }
    script.lappend(
        Tcl_NewStringObj(const_cast<char *>(operation.c_str()), -1));

    // Convert arguments to Tcl values.
    va_list pArg;
    va_start(pArg, pThis);
    const Method::Parameters &parameters = pMethod->parameters();
    Method::Parameters::const_iterator pParam = parameters.begin();
    for (; pParam != parameters.end(); ++pParam) {
        // Append argument value.
        TclObject argument;
        pArg = getArgument(
            pThis->m_threadId, pThis->m_interp, pArg, argument, *pParam);
        script.lappend(argument);
    }

    // Set end of arguments pointer.
    if (pMethod->type().vartype() != VT_VOID) {
        va_arg(pArg, void *);
    }
    pArgEnd = reinterpret_cast<DWORD>(pArg);
    va_end(pArg);

    // Call the Tcl command.
    TclObject result;
    int completionCode = CommandProcessor::instance().eval(
        pThis->m_threadId,
        pThis->m_interp,
        script,
        pThis->m_isSink ? 0 : &result);
    if (completionCode != TCL_OK) {
        hresult = E_UNEXPECTED;
        return;
    }

    // Get values of out parameters.
    va_start(pArg, pThis);
    pParam = parameters.begin();
    for (; pParam != parameters.end(); ++pParam) {
        if (pParam->flags() & PARAMFLAG_FOUT) {
            TclObject varName = getOutVariableName(*pParam);

            // Construct a Tcl command to get the variable value.
            TclObject script(Tcl_NewListObj(0, 0));
            script.lappend(Tcl_NewStringObj("set", -1));
            script.lappend(varName);

            TclObject value;
            CommandProcessor::instance().eval(
                pThis->m_threadId, pThis->m_interp, script, &value);

            pArg = putArgument(pArg, pThis->m_interp, value, pParam->type());
        } else {
            TclObject argument;
            pArg = getArgument(
                pThis->m_threadId, pThis->m_interp, pArg, argument, *pParam);
        }
    }

    // Convert return value.
    if (pMethod->type().vartype() != VT_VOID) {
        putArgument(pArg, pThis->m_interp, result, pMethod->type());
    }

    va_end(pArg);
    hresult = S_OK;
}
