// $Id: ComObjectFactory.cpp,v 1.6 2000/04/26 16:43:18 chuang Exp $
#pragma warning(disable: 4786)
#include "CommandProcessor.h"
#include "ComModule.h"
#include "ComObject.h"
#include "ComObjectFactory.h"

ComObjectFactory::ComObjectFactory (const Interface &interfaceDesc,
                                    Tcl_ThreadId threadId,
                                    Tcl_Interp *interp,
                                    TclObject constructor,
                                    TclObject destructor):
        m_refCount(0),
        m_interfaceDesc(interfaceDesc),
        m_threadId(threadId),
        m_interp(interp),
        m_constructor(constructor),
        m_destructor(destructor)
{ }

void
ComObjectFactory::registerFactory (REFCLSID clsid)
{
    HRESULT hr = CoRegisterClassObject(
        clsid,
        this,
        CLSCTX_SERVER,
        REGCLS_MULTIPLEUSE,
        &m_classObjectHandle);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }
}

void
ComObjectFactory::revokeFactory ()
{
    HRESULT hr = CoRevokeClassObject(m_classObjectHandle);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }
}

STDMETHODIMP
ComObjectFactory::QueryInterface (REFIID iid, void **ppvObj)
{
    if (IsEqualIID(iid, IID_IClassFactory)
     || IsEqualIID(iid, IID_IUnknown)) {
        *ppvObj = this;
        AddRef();
        return S_OK;
    }

    *ppvObj = 0;
    return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG)
ComObjectFactory::AddRef ()
{
    return InterlockedIncrement(&m_refCount);
}

STDMETHODIMP_(ULONG)
ComObjectFactory::Release ()
{
    InterlockedDecrement(&m_refCount);
    if (m_refCount == 0) {
        delete this;
        return 0;
    }
    return m_refCount;
}

STDMETHODIMP
ComObjectFactory::CreateInstance (IUnknown *pOuter, REFIID iid, void **ppvObj)
{
    // We don't support aggregation.
    if (pOuter != 0) {
        *ppvObj = 0;
        return CLASS_E_NOAGGREGATION;
    }

    // Execute Tcl command to create a servant.
    TclObject servant;
    int completionCode = CommandProcessor::instance().eval(
        m_threadId, m_interp, m_constructor, &servant);
    if (completionCode != TCL_OK) {
        *ppvObj = 0;
        return E_UNEXPECTED;
    }

    // Create a COM object and tie its implementation to the servant.
    IUnknown *pComObject = ComObject::newInstance(
        m_interfaceDesc,
        m_threadId,
        m_interp,
        servant,
        m_destructor);

    return pComObject->QueryInterface(iid, ppvObj);
}

STDMETHODIMP
ComObjectFactory::LockServer (BOOL lock)
{
    if (lock) {
        ComModule::instance().lock();
    } else {
        ComModule::instance().unlock();
    }
    return S_OK;
}


SingletonObjectFactory::SingletonObjectFactory (const Interface &interfaceDesc,
                                                Tcl_ThreadId threadId,
                                                Tcl_Interp *interp,
                                                TclObject constructor,
                                                TclObject destructor):
        ComObjectFactory(
            interfaceDesc, threadId, interp, constructor, destructor),
        m_pInstance(0)
{ }

STDMETHODIMP
SingletonObjectFactory::CreateInstance (IUnknown *pOuter,
                                        REFIID iid,
                                        void **ppvObj)
{
    if (m_pInstance == 0) {
	LOCK_MUTEX(m_mutex)
        if (m_pInstance == 0) {
	    HRESULT hr = ComObjectFactory::CreateInstance(
                pOuter,
                iid,
                reinterpret_cast<void **>(&m_pInstance));
            if (FAILED(hr)) {
                return hr;
            }
        }
    }

    return m_pInstance->QueryInterface(iid, ppvObj);
}
