// $Id: HandleSupport.h,v 1.12 2000/04/24 17:08:45 chuang Exp $
#ifndef HANDLESUPPORT_H
#define HANDLESUPPORT_H

#pragma warning(disable: 4786)
#include <tcl.h>
#include "StringHashTable.h"

extern void handleObjDupInternalRep(Tcl_Obj *pSrc, Tcl_Obj *pDup);
extern void handleObjUpdateString(Tcl_Obj *pObj);

// This class manages the life cycle of handle objects that represent
// application objects.  Handles are instances of a new Tcl object type.
// The internal representation of this Tcl object type is a pointer to an
// application object.  So given a handle object in this form, it's a trivial
// operation to get the application object represented by the handle.
//
// Some Tcl commands may convert the handle object to another object type.
// To get the application object in this case, this class must map from the
// string representation of the Tcl object to the application object.

template<class RepType>
class HandleSupport
{
    // Tcl object type used to represent handles
    static Tcl_ObjType m_objType;

    // Tcl interpreter that creates handles
    Tcl_Interp *m_interp;

    // function that implements the object command
    Tcl_ObjCmdProc *m_pCmdProc;

    // handle to object representation map
    typedef StringHashTable<RepType *> Map;
    Map m_map;

    // Insert handle to object mapping.
    void insert (Tcl_Obj *pHandle, RepType *pRep)
    { m_map.insert(Tcl_GetStringFromObj(pHandle, 0), pRep); }

    // Get handle from string representation.
    static int handleObjSetFromAny(Tcl_Interp *interp, Tcl_Obj *pObj);
    int setFromAny(Tcl_Interp *interp, Tcl_Obj *pObj);

    // Called when the Tcl interpreter is deleted.
    static void interpDeleteProc (ClientData clientData, Tcl_Interp *)
    { delete reinterpret_cast<HandleSupport<RepType> *>(clientData); }

public:
    HandleSupport(
        Tcl_Interp *interp, const char *name, Tcl_ObjCmdProc *pCmdProc);

    // Get instance of the class for the given Tcl intepreter.
    static HandleSupport<RepType> &instance(Tcl_Interp *interp);

    // Create a handle and associate it with an application object.
    // The caller is responsible for the life cycle of the application object.
    Tcl_Obj *newObj(RepType *pRep);

    // Get count of matching elements.
    size_t count (Tcl_Obj *pHandle) const
    { return (m_map.find(Tcl_GetStringFromObj(pHandle, 0)) == 0) ? 0 : 1; }

    // Get the object represented by the handle.  If the handle is invalid,
    // put an error message in the Tcl interpreter result and return 0.
    RepType *find(Tcl_Obj *pHandle) const;

    // Destroy handle.
    void erase(Tcl_Obj *pHandle);
};

template<class RepType>
Tcl_ObjType HandleSupport<RepType>::m_objType;

template<class RepType>
HandleSupport<RepType>::HandleSupport (Tcl_Interp *interp,
                                       const char *name,
                                       Tcl_ObjCmdProc *pCmdProc):
    m_interp(interp),
    m_pCmdProc(pCmdProc)
{
    m_objType.name = const_cast<char *>(name);
    m_objType.freeIntRepProc = NULL;
    m_objType.dupIntRepProc = handleObjDupInternalRep;
    m_objType.updateStringProc = handleObjUpdateString;
    m_objType.setFromAnyProc = handleObjSetFromAny;
    Tcl_RegisterObjType(&m_objType);

    Tcl_SetAssocData(
        interp,
        m_objType.name,
        interpDeleteProc,
        reinterpret_cast<ClientData>(this));
}

template<class RepType>
int
HandleSupport<RepType>::handleObjSetFromAny (Tcl_Interp *interp,
                                             Tcl_Obj *pObj)
{
    return instance(interp).setFromAny(interp, pObj);
}

template<class RepType>
int
HandleSupport<RepType>::setFromAny (Tcl_Interp *interp, Tcl_Obj *pObj)
{
    // Check if the string represents an existing handle.
    char *handleStr = Tcl_GetStringFromObj(pObj, 0);
    RepType *pRep = m_map.find(handleStr);
    if (pRep == 0) {
        if (interp != NULL) {
            Tcl_AppendResult(interp, "invalid handle ", handleStr, NULL);
        }
        return TCL_ERROR;
    }

    // Free the old internal representation.
    if (pObj->typePtr != NULL && pObj->typePtr->freeIntRepProc != NULL) {
        pObj->typePtr->freeIntRepProc(pObj);
    }

    pObj->typePtr = &m_objType;
    pObj->internalRep.otherValuePtr = pRep;
    return TCL_OK;
}

template<class RepType>
HandleSupport<RepType> &
HandleSupport<RepType>::instance (Tcl_Interp *interp)
{
    return *reinterpret_cast<HandleSupport<RepType> *>(
        Tcl_GetAssocData(interp, m_objType.name, 0));
}

template<class RepType>
RepType *
HandleSupport<RepType>::find (Tcl_Obj *pObj) const
{
    // This function call should return immediately if the object is already of
    // the desired type.  If it's not, the conversion function tries to map
    // the string representation to a pointer to a RepType object.
    if (Tcl_ConvertToType(m_interp, pObj, const_cast<Tcl_ObjType *>(&m_objType))
     != TCL_OK) {
        return 0;
    }
    return reinterpret_cast<RepType *>(pObj->internalRep.otherValuePtr);
}

template<class RepType>
void
HandleSupport<RepType>::erase (Tcl_Obj *pObj)
{
    char *handleStr = Tcl_GetStringFromObj(pObj, 0);
    if (m_map.find(handleStr) != 0) {
        Tcl_DeleteCommand(m_interp, handleStr);
        m_map.erase(handleStr);
    }
}

template<class RepType>
Tcl_Obj *
HandleSupport<RepType>::newObj (RepType *pRep)
{
    // Create a handle.
    Tcl_Obj *pObj = Tcl_NewObj();
    Tcl_InvalidateStringRep(pObj);
    pObj->typePtr = &m_objType;
    pObj->internalRep.otherValuePtr = pRep;

    Tcl_CreateObjCommand(
        m_interp,
        Tcl_GetStringFromObj(pObj, 0),
        m_pCmdProc,
        reinterpret_cast<ClientData>(pRep),
        0);

    insert(pObj, pRep);
    return pObj;
}

#endif
