// $Id: TypeInfo.cpp,v 1.34 2000/05/09 15:41:03 chuang Exp $
#pragma warning(disable: 4786)
#include <sstream>
#include <map>
#include "Uuid.h"
#include "TypeInfo.h"

// This class maps from a string representation to a VARTYPE.

class StringToVarTypeMap: public std::map<std::string, int>
{
public:
    StringToVarTypeMap();
};

StringToVarTypeMap::StringToVarTypeMap ()
{
    insert(value_type("EMPTY", VT_EMPTY));
    insert(value_type("NULL", VT_NULL));
    insert(value_type("I2", VT_I2));
    insert(value_type("I4", VT_I4));
    insert(value_type("R4", VT_R4));
    insert(value_type("R8", VT_R8));
    insert(value_type("CY", VT_CY));
    insert(value_type("DATE", VT_DATE));
    insert(value_type("BSTR", VT_BSTR));
    insert(value_type("DISPATCH", VT_DISPATCH));
    insert(value_type("ERROR", VT_ERROR));
    insert(value_type("BOOL", VT_BOOL));
    insert(value_type("VARIANT", VT_VARIANT));
    insert(value_type("UNKNOWN", VT_UNKNOWN));
    insert(value_type("DECIMAL", VT_DECIMAL));
    insert(value_type("RECORD", VT_RECORD));
    insert(value_type("I1", VT_I1));
    insert(value_type("UI1", VT_UI1));
    insert(value_type("UI2", VT_UI2));
    insert(value_type("UI4", VT_UI4));
    insert(value_type("INT", VT_INT));
    insert(value_type("UINT", VT_UINT));
    insert(value_type("ARRAY", VT_ARRAY));
    insert(value_type("VOID", VT_VOID));
}

static StringToVarTypeMap stringToVarTypeMap;

Type::Type (const std::string &str):
    m_pointerCount(0)
{
    std::istringstream in(str);
    std::string token;
    while (in >> token) {
        if (token[0] == '*') {
            ++m_pointerCount;
        } else {
            StringToVarTypeMap::iterator p = stringToVarTypeMap.find(token);
            if (p == stringToVarTypeMap.end()) {
                m_vt = VT_USERDEFINED;
                m_name = token;
            } else {
                m_vt = p->second;
            }
        }
    }
}

std::string
Type::toString () const
{
    std::string name;

    switch (m_vt) {
    case VT_EMPTY:
        name = "EMPTY";
        break;
    case VT_NULL:
        name = "NULL";
        break;
    case VT_I2:
        name = "I2";
        break;
    case VT_I4:
        name = "I4";
        break;
    case VT_R4:
        name = "R4";
        break;
    case VT_R8:
        name = "R8";
        break;
    case VT_CY:
        name = "CY";
        break;
    case VT_DATE:
        name = "DATE";
        break;
    case VT_BSTR:
        name = "BSTR";
        break;
    case VT_DISPATCH:
        name = "DISPATCH";
        break;
    case VT_ERROR:
        name = "ERROR";
        break;
    case VT_BOOL:
        name = "BOOL";
        break;
    case VT_VARIANT:
        name = "VARIANT";
        break;
    case VT_UNKNOWN:
        name = "UNKNOWN";
        break;
    case VT_DECIMAL:
        name = "DECIMAL";
        break;
    case VT_RECORD:
        name = "RECORD";
        break;
    case VT_I1:
        name = "I1";
        break;
    case VT_UI1:
        name = "UI1";
        break;
    case VT_UI2:
        name = "UI2";
        break;
    case VT_UI4:
        name = "UI4";
        break;
    case VT_INT:
        name = "INT";
        break;
    case VT_UINT:
        name = "UINT";
        break;
    case VT_ARRAY:
        name = "ARRAY";
        break;
    case VT_BYREF:
        name = "BYREF";
        break;
    case VT_VOID:
        name = "VOID";
        break;
    case VT_USERDEFINED:
        name = m_name;
        break;
    default:
        std::ostringstream out;
        out << "unknown0x" << std::hex << m_vt;
        name = out.str();
    }
    return name;
}

void
Type::readTypeDesc (TypeInfo &typeInfo, TYPEDESC &typeDesc)
{
    HRESULT hr;

    switch (typeDesc.vt) {
    case VT_USERDEFINED:
	// It's an alias: Expand the alias.
        ITypeInfo *pITypeInfoRef;
	hr = typeInfo->GetRefTypeInfo(typeDesc.hreftype, &pITypeInfoRef);
        if (SUCCEEDED(hr)) {
            TypeInfo typeInfoRef(pITypeInfoRef);

	    if (m_name.empty()) {
                m_name = typeInfoRef.name();
	    }

            TypeAttr typeAttr(typeInfoRef);
            if (typeAttr->typekind == TKIND_ALIAS) {
		// Type expanded to another alias!
		readTypeDesc(typeInfoRef, typeAttr->tdescAlias);
            } else {
        	m_vt = typeDesc.vt;
                m_iid = typeAttr->guid;
            }
	}
	break;

    case VT_PTR:
	// It's a pointer: Dereference and try to interpret with one less
	// level of indirection.
	++m_pointerCount;
        readTypeDesc(typeInfo, *typeDesc.lptdesc);
	break;

    default:
	m_vt = typeDesc.vt;
    }
}

Parameter::Parameter (TypeInfo &typeInfo,
                      ELEMDESC &elemDesc,
                      const char *name):
    m_flags(elemDesc.paramdesc.wParamFlags),
    m_type(typeInfo, elemDesc.tdesc),
    m_name(name)
{
    if (m_flags == 0) {
        // No parameter passing flags are set.  Assume it's an in parameter.
        m_flags = PARAMFLAG_FIN;
    }
}

Parameter::Parameter (const std::string &modes,
                      const std::string &type,
                      const std::string &name):
    m_flags(0),
    m_type(type),
    m_name(name)
{
    std::istringstream in(modes);
    std::string token;
    while (in >> token) {
        if (token == "in") {
            m_flags |= PARAMFLAG_FIN;
        } else if (token == "out") {
            m_flags |= PARAMFLAG_FOUT;
        }
    }
}


Method::Method (TypeInfo &typeInfo, DISPID memberid, ELEMDESC &elemDesc):
    m_memberid(memberid),
    m_type(typeInfo, elemDesc.tdesc)
{
    // Get name.
    BSTR nameBstr;
    unsigned numNames;
    HRESULT hr = typeInfo->GetNames(
        memberid,
        &nameBstr,
        1,
        &numNames);
    if (FAILED(hr)) {
        // TODO: I should throw an exception here.
        return;
    }

    // Initialize name.
    _bstr_t name(nameBstr, false);
    m_name = std::string(name);
}

Method::Method (TypeInfo &typeInfo, FuncDesc &funcDesc):
    m_memberid(funcDesc->memid),
    m_type(typeInfo, funcDesc->elemdescFunc.tdesc),
    m_invokeKind(funcDesc->invkind),
    m_vtblIndex(funcDesc->oVft / 4)
{
    // Get method and parameter names.
    BSTR *pNames = new BSTR[funcDesc->cParams + 1];
    unsigned numNames;
    HRESULT hr = typeInfo->GetNames(
        funcDesc->memid,
        pNames,
        funcDesc->cParams + 1,
        &numNames);
    if (FAILED(hr)) {
        // TODO: I should throw an exception here.
        delete[] pNames;
        return;
    }

    // Initialize method name.
    _bstr_t methodName(pNames[0], false);
    m_name = std::string(methodName);

    // Get parameter types.
    for (unsigned i = 1; i < numNames; ++i) {
        _bstr_t paramName(pNames[i], false);
        Parameter parameter(
            typeInfo,
            funcDesc->lprgelemdescParam[i - 1],
            paramName);
        addParameter(parameter);
    }

    // Some TKIND_INTERFACE methods specify a parameter is a return value
    // with the PARAMFLAG_FRETVAL flag.  Check if the last parameter is
    // actually a return value.
    if (m_type.vartype() == VT_HRESULT) {
        m_type = Type("VOID");

        if (m_parameters.size() > 0) {
            Parameter last = m_parameters.back();
            if ((last.flags() & (PARAMFLAG_FOUT|PARAMFLAG_FRETVAL))
             == (PARAMFLAG_FOUT|PARAMFLAG_FRETVAL)) {
                m_type = last.type();
                m_parameters.pop_back();
            }
        }
    }

    delete[] pNames;
}

Method::Method (MEMBERID memberid,
                const std::string &type,
                const std::string &name,
                INVOKEKIND invokeKind):
    m_memberid(memberid),
    m_type(type),
    m_name(name),
    m_invokeKind(invokeKind)
{ }


Property::Property (TypeInfo &typeInfo, FuncDesc &funcDesc):
    Method(typeInfo, funcDesc),
    m_readOnly(true)
{ }

Property::Property (TypeInfo &typeInfo, VarDesc &varDesc):
    Method(typeInfo, varDesc->memid, varDesc->elemdescVar),
    m_readOnly((varDesc->wVarFlags & VARFLAG_FREADONLY) != 0)
{ }

Property::Property (MEMBERID memberid,
                    const std::string &modes,
                    const std::string &type,
                    const std::string &name):
    Method(memberid, type, name, INVOKE_PROPERTYGET),
    m_readOnly(true)
{
    // Initialize readable/writable flags.
    std::istringstream in(modes);
    std::string token;
    while (in >> token) {
        if (token == "in") {
            m_readOnly = false;
        }
    }
}


std::string
Interface::iidString () const
{
    Uuid uuid(m_iid);
    return uuid.toString();
}

void
Interface::addMethod (const Method &method)
{
    // Do we already have information on this method?
    Methods::iterator p = m_methods.begin();
    for (; p != m_methods.end(); ++p) {
        if (p->memberid() == method.memberid()
         && p->vtblIndex() == method.vtblIndex()) {
            return;
        }
    }

    // Add method description.
    m_methods.push_back(method);
}

void
Interface::addProperty (const Property &property)
{
    // Do we already have information on this property?
    Properties::iterator p = m_properties.begin();
    for (; p != m_properties.end(); ++p) {
        if (p->memberid() == property.memberid()) {
            // Set read only flag.
            if (property.invokeKind() != INVOKE_PROPERTYGET) {
                p->readOnly(false);
            }

            // Set property value type.
            if (property.type().vartype() != VT_VOID) {
                p->type(property.type());
            }
            return;
        }
    }

    // Add property description.
    m_properties.push_back(property);
}

void
Interface::readFunctions (TypeInfo &typeInfo, TypeAttr &typeAttr)
{
    HRESULT hr;

    // Get properties and methods from inherited interfaces.
    for (unsigned i = 0; i < typeAttr->cImplTypes; ++i) {
	HREFTYPE hRefType;
	hr = typeInfo->GetRefTypeOfImplType(i, &hRefType);
        if (FAILED(hr)) {
            break;
        }

        ITypeInfo *pTypeInfo;
	hr = typeInfo->GetRefTypeInfo(hRefType, &pTypeInfo);
        if (FAILED(hr)) {
            break;
        }
        TypeInfo superTypeInfo(pTypeInfo);
        TypeAttr superTypeAttr(superTypeInfo);
        readFunctions(superTypeInfo, superTypeAttr);
    }

    // Get properties and methods of this interface.
    for (unsigned j = 0; j < typeAttr->cFuncs; ++j) {
        FuncDesc funcDesc(typeInfo, j);

#if 0
	if ((typeAttr->wTypeFlags & TYPEFLAG_FDUAL)
	 && funcDesc->oVft < 28) {
	    // Don't expose the first seven functions in a dual interface
	    // to the Tcl script because of potential dangers. These are:
	    //	IUnknown::QueryInterface
	    //	IUnknown::AddRef
	    //	IUnknown::Release
	    //	IDispatch::GetTypeInfoCount
	    //	IDispatch::GetTypeInfo
	    //	IDispatch::GetIDsOfNames
	    //	IDispatch::Invoke
	    continue;
	}
#endif

        Method method(typeInfo, funcDesc);
        addMethod(method);

	if (funcDesc->invkind != INVOKE_FUNC) {
	    // This is a property get/set function.
            Property property(typeInfo, funcDesc);
            addProperty(property);
	}
    }

    // Some objects expose their properties as variable members.
    for (unsigned k = 0; k < typeAttr->cVars; ++k) {
        VarDesc varDesc(typeInfo, k);

	if (varDesc->varkind == VAR_DISPATCH) {
            Property property(typeInfo, varDesc);
            addProperty(property);
	}
    }

    // Add missing parameter description to property put functions.
    Methods::iterator pMethod = m_methods.begin();
    for (; pMethod != m_methods.end(); ++pMethod) {
        if (pMethod->invokeKind() == INVOKE_PROPERTYPUT
         || pMethod->invokeKind() == INVOKE_PROPERTYPUTREF) {
            const Property *pProperty = findProperty(pMethod->name().c_str());
            if (pProperty != 0) {
                Parameter parameter(
                    PARAMFLAG_FIN,
                    pProperty->type(),
                    "propertyValue");
                pMethod->addParameter(parameter);
            }
        }
    }
}

void
Interface::readTypeInfo (TypeInfo &typeInfo)
{
    m_name = typeInfo.name();

    TypeAttr typeAttr(typeInfo);
    m_iid = typeAttr->guid;

    m_dispatchable = (typeAttr->typekind == TKIND_DISPATCH
        || (typeAttr->wTypeFlags & TYPEFLAG_FDISPATCHABLE) != 0);

    readFunctions(typeInfo, typeAttr);
}

const Method *
Interface::findMethod (const char *name) const
{
    Methods::const_iterator p = m_methods.begin();
    for (; p != m_methods.end(); ++p) {
        if (p->name() == name) {
            return &(*p);
        }
    }
    return 0;
}

const Property *
Interface::findProperty (const char *name) const
{
    Properties::const_iterator p = m_properties.begin();
    for (; p != m_properties.end(); ++p) {
        if (p->name() == name) {
            return &(*p);
        }
    }
    return 0;
}


Singleton<InterfaceManager> InterfaceManager::ms_singleton;

InterfaceManager &
InterfaceManager::instance ()
{
    return ms_singleton.instance();
}

InterfaceManager::InterfaceManager ()
{
}

InterfaceManager::~InterfaceManager ()
{
}

const Interface *
InterfaceManager::newInterface (REFIID iid, TypeInfo &typeInfo)
{
    Interface *pInterface = m_hashTable.find(iid);
    if (pInterface == 0) {
        pInterface = new Interface(typeInfo);
        m_hashTable.insert(iid, pInterface);
    }
    return pInterface;
}

Interface *
InterfaceManager::newInterface (REFIID iid, const char *name)
{
    Interface *pInterface = m_hashTable.find(iid);
    if (pInterface == 0) {
        pInterface = new Interface(iid, name);
        m_hashTable.insert(iid, pInterface);
    }
    return pInterface;
}

const Interface *
InterfaceManager::find (REFIID iid) const
{
    return m_hashTable.find(iid);
}


std::string
TypeInfo::name (MEMBERID id)
{
    BSTR nameStr = 0;
    HRESULT hResult = m_pTypeInfo->GetDocumentation(
        id, &nameStr, NULL, NULL, NULL);
    if (FAILED(hResult) || nameStr == 0) {
        return std::string();
    }
    _bstr_t wrapper(nameStr, false);
    return std::string(wrapper);
}

ITypeInfo *
TypeInfo::getImplType (int bits)
{
    HRESULT hr;

    TypeAttr classAttr(*this);
    for (unsigned i = 0; i < classAttr->cImplTypes; ++i) {
	int flags;
	hr = m_pTypeInfo->GetImplTypeFlags(i, &flags);
        if (FAILED(hr)) {
	    break;
	}

        const int MASK = IMPLTYPEFLAG_FDEFAULT | IMPLTYPEFLAG_FSOURCE
            | IMPLTYPEFLAG_FRESTRICTED;
	if ((flags & MASK) != bits) {
	    continue;
	}

	HREFTYPE hRefType;
	hr = m_pTypeInfo->GetRefTypeOfImplType(i, &hRefType);
        if (FAILED(hr)) {
            break;
        }

        ITypeInfo *pTypeInfo;
	hr = m_pTypeInfo->GetRefTypeInfo(hRefType, &pTypeInfo);
        if (FAILED(hr)) {
            break;
        }
        return pTypeInfo;
    }

    return 0;
}

FuncDesc::FuncDesc (TypeInfo &typeInfo, unsigned index):
    m_typeInfo(typeInfo)
{
    HRESULT hr = m_typeInfo->GetFuncDesc(index, &m_pFuncDesc);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }
}

VarDesc::VarDesc (TypeInfo &typeInfo, unsigned index):
    m_typeInfo(typeInfo)
{
    HRESULT hr = m_typeInfo->GetVarDesc(index, &m_pVarDesc);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }
}
