// $Id: refCmd.cpp,v 1.12 2000/04/20 18:37:40 chuang Exp $
#include <sstream>
#include "TypeInfo.h"
#include "TclObject.h"
#include "ComInitializer.h"
#include "refCmd.h"

int
setComErrorResult (Tcl_Interp *interp,
                   _com_error &e,
                   const char *file,
                   int line)
{
    TclObject result(Tcl_NewListObj(0, 0));

    // Append HRESULT value in hexadecimal string format.
    std::ostringstream hr;
    hr << "0x" << std::hex << e.Error();
    TclObject hrObj(hr.str());
    result.lappend(hrObj);

    // Append error message.
#if TCL_MINOR_VERSION >= 2
    // Uses Unicode functions introduced in Tcl 8.2.
    Tcl_UniChar *pMessage = 0;
    FormatMessageW(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        e.Error(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        reinterpret_cast<LPWSTR>(&pMessage),
        0,
        NULL);

    if (pMessage != 0) {
        int nLen = wcslen(pMessage);
        if (nLen > 1 && pMessage[nLen - 1] == '\n') {
            --nLen;
            if (nLen > 1 && pMessage[nLen - 1] == '\r') {
                --nLen;
            }
        }
        result.lappend(Tcl_NewUnicodeObj(pMessage, nLen));
    } else {
        pMessage = reinterpret_cast<Tcl_UniChar *>(
            LocalAlloc(0, 32 * sizeof(WCHAR)));
        if (pMessage != 0) {
            WORD wCode = e.WCode();
            if (wCode != 0) {
                swprintf(pMessage, L"IDispatch error #%d", wCode);
            } else {
                swprintf(pMessage, L"Unknown error 0x%0lX", e.Error());
            }
            result.lappend(Tcl_NewUnicodeObj(pMessage, -1));
        }
    }
    LocalFree(pMessage);
#else
    result.lappend(Tcl_NewStringObj(e.ErrorMessage(), -1));
#endif

#ifndef NDEBUG
    // Append file and line number.
    std::ostringstream fileLine;
    fileLine << file << ' ' << line;
    TclObject fileLineObj(fileLine.str());
    result.lappend(fileLineObj);
#endif

    Tcl_SetObjResult(interp, result);
    return TCL_ERROR;
}

// Invoke a method or property.

static int
invoke (Tcl_Interp *interp,
        int objc,               // number of arguments
        Tcl_Obj *CONST objv[],  // arguments
        Reference *pReference,
        const Method *pMethod,
        Arguments &arguments,
        INVOKEKIND invokeKind)
{
    // Set up return value.
    _variant_t returnValue;
    _variant_t *pReturnValue = (pMethod->type().vartype() == VT_VOID)
        ? 0 : &returnValue;

    // Invoke it.
    HRESULT hr;
    hr = pReference->invoke(
        pMethod->memberid(),
        invokeKind,
        arguments,
        pReturnValue);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    // Store values returned from out parameters.
    arguments.storeOutValues(interp, objc, objv, pMethod->parameters());

    // Convert return value.
    if (pReturnValue != 0) {
        TclObject value(returnValue, pMethod->type(), interp);
        Tcl_SetObjResult(interp, value);
    }
    return TCL_OK;
}

// Set Tcl result to a wrong number of arguments error message.

static void
wrongNumArgs (
    Tcl_Interp *interp,		            // current interpreter
    Tcl_Obj *CONST objv[],                  // method name
    const Method::Parameters &parameters)   // expected parameters
{
    bool first = true;
    std::ostringstream paramNames;
    for (Property::Parameters::const_iterator p =
     parameters.begin(); p != parameters.end(); ++p) {
        if (first) {
            first = false;
        } else {
            paramNames << ' ';
        }
        paramNames << p->name();
    }

    Tcl_WrongNumArgs(
        interp, 1, objv, const_cast<char *>(paramNames.str().c_str()));
}

// Get or put an object property.

static int
invokeProperty (
    Tcl_Interp *interp,		// Current interpreter
    int objc,
    Tcl_Obj *CONST objv[],	// property name and arguments
    Reference *pReference,
    const Property *pProperty)
{
    INVOKEKIND invokeKind;
    const Property::Parameters &parameters = pProperty->parameters();

    if (objc == parameters.size() + 2) {
        // Put property.
        invokeKind = INVOKE_PROPERTYPUT;

    } else if (objc == parameters.size() + 1) {
        // Get property.
        invokeKind = INVOKE_PROPERTYGET;

    } else {
	wrongNumArgs(interp, objv, parameters);
        return TCL_ERROR;
    }

    PositionalArguments arguments;
    int result = arguments.init(
        interp, objc - 1, objv + 1, *pProperty, invokeKind);
    if (result != TCL_OK) {
        return result;
    }

    return invoke(
        interp,
        objc - 1,
        objv + 1,
        pReference,
        pProperty,
        arguments,
        invokeKind);
}

// This Tcl command invokes a method or property on an interface pointer.

int
referenceObjCmd (
    ClientData clientData,	/* Client data. */
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* The argument objects. */
{
    bool namedArgOpt = false;

    int i = 1;
    for (; i < objc; ++i) {
        static char *options[] = {
	    "-namedarg", "-property", NULL
        };
        enum OptionEnum {
            OPTION_NAMEDARG, OPTION_PROPERTY
        };

        int index;
        if (Tcl_GetIndexFromObj(NULL, objv[i], options, "option", 0, &index)
         != TCL_OK) {
            break;
        }

        switch (index) {
        case OPTION_NAMEDARG:
            namedArgOpt = true;
            break;
        }
    }

    if (objc - i < 1) {
	Tcl_AppendResult(
            interp, "usage: handle ?options? method ?arg ...?", NULL);
	return TCL_ERROR;
    }

    Reference *pReference = reinterpret_cast<Reference *>(clientData);

    int result;
    try {
        // Get interface description.
        const Interface *pInterface = pReference->interfaceDesc();
        if (pInterface == 0) {
            Tcl_AppendResult(interp, "no interface information", NULL);
            return TCL_ERROR;
        }

        const Method *pMethod;
        const Property *pProperty;
        char *name = Tcl_GetStringFromObj(objv[i], (int *)0);

        if ((pProperty = pInterface->findProperty(name)) != 0) {
            // It's a property.
            result = invokeProperty(
                interp,
                objc - i,
                objv + i,
                pReference,
                pProperty);

        } else if ((pMethod = pInterface->findMethod(name)) != 0) {
            // It's a method.
            ++i;
            NamedArguments namedArguments;
            PositionalArguments positionalArguments;
            Arguments *pArguments;

            if (namedArgOpt) {
                pArguments = &namedArguments;
            } else {
                // Return an error if too many arguments were given.
                const Method::Parameters &parameters = pMethod->parameters();
                if (objc - i > parameters.size()) {
                    wrongNumArgs(interp, objv, parameters);
                    return TCL_ERROR;
                }
                pArguments = &positionalArguments;
            }
            result = pArguments->init(
                interp, objc - i, objv + i, *pMethod, INVOKE_FUNC);
            if (result != TCL_OK) {
                return result;
            }

            result = invoke(
                interp,
                objc - i,
                objv + i,
                pReference,
                pMethod,
                *pArguments,
                INVOKE_FUNC);

        } else {
            Tcl_AppendResult(
                interp, "unknown method or property ", name, NULL);
            result = TCL_ERROR;
        }
    }
    catch (_com_error &e) {
        result = setComErrorResult(interp, e, __FILE__, __LINE__);
    }
    return result;
}

// This command queries an interface pointer for a given interface.

static int
queryInterfaceCmd (
    Tcl_Interp *interp,
    int objc,
    Tcl_Obj *CONST objv[])
{
    if (objc != 4) {
	Tcl_WrongNumArgs(interp, 2, objv, "handle IID");
	return TCL_ERROR;
    }

    Reference *pReference =
        HandleSupport<Reference>::instance(interp).find(objv[2]);
    if (pReference == 0) {
        char *arg = Tcl_GetStringFromObj(objv[2], (int *)0);
        Tcl_AppendResult(interp, "invalid interface pointer handle ", arg, NULL);
        return TCL_ERROR;
    }

    char *iidStr = Tcl_GetStringFromObj(objv[3], (int *)0);
    IID iid;
    if (UuidFromString(reinterpret_cast<unsigned char *>(iidStr), &iid)
     != RPC_S_OK) {
	Tcl_AppendResult(
            interp,
            "cannot convert to IID: ",
            iidStr,
            NULL);
        return TCL_ERROR;
    }

    try {
        Reference *pNewRef = ReferenceFactory::instance().newReference(
            pReference->dispatch(), iid);
        Tcl_SetObjResult(
            interp,
            HandleSupport<Reference>::instance(interp).newObj(pNewRef));
    }
    catch (_com_error &e) {
        return setComErrorResult(interp, e, __FILE__, __LINE__);
    }
    return TCL_OK;
}

// This Tcl command gets a reference to an object.

int
refCmd (
    ClientData clientData,	/* Not used. */
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* The argument objects. */
{
    ComInitializer::initialize();

    if (objc < 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "subcommand argument ...");
	return TCL_ERROR;
    }

    static char *options[] = {
	"createobject", "getobject", "queryinterface", NULL
    };
    enum SubCommandEnum {
        CREATEOBJECT, GETOBJECT, QUERYINTERFACE
    };

    int subCommand;
    if (Tcl_GetIndexFromObj(interp, objv[1], options, "subcommand", 0,
     &subCommand) != TCL_OK) {
	return TCL_ERROR;
    }

    if (subCommand == QUERYINTERFACE) {
        return queryInterfaceCmd(interp, objc, objv);
    }

    bool clsIdOpt = false;

    int i = 2;
    for (; i < objc; ++i) {
        static char *options[] = {
	    "-clsid", NULL
        };
        enum OptionEnum {
            OPTION_CLSID
        };

        int index;
        if (Tcl_GetIndexFromObj(NULL, objv[i], options, "option", 0, &index)
         != TCL_OK) {
            break;
        }

        switch (index) {
        case OPTION_CLSID:
            clsIdOpt = true;
            break;
        }
    }

    char *progId = Tcl_GetStringFromObj(objv[i], (int *)0);
    char *hostName = (i + 1 < objc)
        ? Tcl_GetStringFromObj(objv[i+1], (int *)0) : 0;

    try {
        Reference *pReference;
        
        if (clsIdOpt) {
            CLSID clsid;
            if (UuidFromString(reinterpret_cast<unsigned char *>(progId), &clsid)
             != RPC_S_OK) {
	        Tcl_AppendResult(
                    interp,
                    "cannot convert to CLSID: ",
                    progId,
                    NULL);
                return TCL_ERROR;
            }
            pReference = (subCommand == GETOBJECT)
                ? ReferenceFactory::instance().getObject(clsid, 0)
                : ReferenceFactory::instance().createObject(
                    clsid, 0, hostName);

        } else {
            pReference = (subCommand == GETOBJECT)
                ? ReferenceFactory::instance().getObject(progId)
                : ReferenceFactory::instance().createObject(
                    progId, hostName);
        }

        Tcl_SetObjResult(
            interp,
            HandleSupport<Reference>::instance(interp).newObj(pReference));
    }
    catch (_com_error &e) {
        return setComErrorResult(interp, e, __FILE__, __LINE__);
    }
    return TCL_OK;
}

#if 0
// This Tcl command gets and puts an object property.

int
propCmd (
    ClientData clientData,	/* Not used. */
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* The argument objects. */
{
    if (objc < 3) {
	Tcl_WrongNumArgs(
            interp,
            1,
            objv,
            "object propName ?index ...? ?value?");
	return TCL_ERROR;
    }

    // Find reference from handle.
    Reference *pReference = referenceHandleSupport.find(interp, objv[1]);
    if (pReference == 0) {
	return TCL_ERROR;
    }

    int result;
    try {
        // Get interface description.
        const Interface *pInterface = pReference->interfaceDesc();
        if (pInterface == 0) {
            Tcl_AppendResult(interp, "no interface information", NULL);
            return TCL_ERROR;
        }

        // Get property description.
        char *name = Tcl_GetStringFromObj(objv[2], (int *)0);
        const Property *pProperty = pInterface->findProperty(name);
        if (pProperty == 0) {
            Tcl_AppendResult(interp, "unknown property ", name, NULL);
            return TCL_ERROR;
        }

        result = invokeProperty(
            interp,
            pReference,
            pProperty,
            objc - 2,
            objv + 2);
    }
    catch (_com_error &e) {
        result = setComErrorResult(interp, e, __FILE__, __LINE__);
    }
    return result;
}
#endif
